import numpy as np
import pandas as pd

import starfile
from .constants import test_data_directory

million_row_file = test_data_directory / '1m_row.star'


def generate_large_star_file():
    df = pd.DataFrame(np.random.randint(0, 100, size=(100000, 4)), columns=list('ABCD'))
    starfile.write(df, million_row_file, overwrite=True)


def remove_large_star_file():
    million_row_file.unlink()



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Union

if TYPE_CHECKING:
    import pandas as pd
    from os import PathLike

from .parser import StarParser
from .writer import StarWriter

if TYPE_CHECKING:
    import pandas as pd
    from os import PathLike


def read(filename: PathLike, read_n_blocks: int = None, always_dict: bool = False):
    """
    Read a star file into a pandas dataframe or dict of pandas dataframes

    default behaviour in the case of only one data block being present in the STAR file is to
    return only a dataframe, this can be changed by setting 'always_dict=True'
    """
    star = StarParser(filename, read_n_blocks=read_n_blocks)
    if len(star.dataframes) == 1 and always_dict is False:
        return star.first_dataframe
    else:
        return star.dataframes


def write(data: Union[pd.DataFrame, Dict[str, pd.DataFrame], List[pd.DataFrame]],
          filename: PathLike,
          float_format: str = '%.6f', sep: str = '\t', na_rep: str = '<NA>',
          overwrite: bool = False, force_loop: bool = True):
    """
    Write dataframes from pandas dataframe(s) to a star file

    data can be a single dataframe, a list of dataframes or a dict of dataframes
    float format defaults to 6 digits after the decimal point
    """
    StarWriter(data, filename=filename, float_format=float_format, overwrite=overwrite,
               na_rep=na_rep, sep=sep, force_loop=force_loop)



from pathlib import Path

import pandas as pd
import pytest
import starfile

from .constants import loop_simple, postprocess, test_df, test_data_directory


def test_read():
    df = starfile.read(loop_simple)
    assert isinstance(df, pd.DataFrame)


def test_read_always_dict():
    data = starfile.read(loop_simple, always_dict=True)
    assert isinstance(data, dict)


def test_read_multiblock():
    data = starfile.read(postprocess)
    assert isinstance(data, dict)
    assert len(data) == 3


def test_write():
    output_file = test_data_directory / 'test_write.star'
    starfile.write(test_df, output_file, overwrite=True)
    assert output_file.exists()


def test_write_fails_to_overwrite_without_flag():
    output_file = test_data_directory / 'test_overwrite_flag.star'
    starfile.write(test_df, output_file, overwrite=True)

    assert output_file.exists()
    with pytest.raises(FileExistsError):
        starfile.write(test_df, output_file, overwrite=False)
        starfile.new(test_df, output_file)


def test_write_overwrites_with_flag():
    output_file = test_data_directory / 'test_overwrite_flag.star'
    starfile.write(test_df, output_file, overwrite=True)

    assert output_file.exists()
    starfile.write(test_df, output_file, overwrite=True)


def test_write_with_float_format():
    output_file = test_data_directory / 'test_write_with_float_format.star'
    test_df['float_col'] = 1.23456789
    starfile.write(test_df, output_file, float_format='%.3f', overwrite=True)
    assert output_file.exists()


def test_read_non_existent_file():
    f = Path('non-existent-file.star')
    assert f.exists() is False

    with pytest.raises(FileNotFoundError):
        starfile.read(f)



from __future__ import annotations

from collections import OrderedDict
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Union, Dict, List

import pandas as pd
from pkg_resources import get_distribution

from .utils import TextBuffer

if TYPE_CHECKING:
    from os import PathLike

__version__ = get_distribution("starfile").version


class StarWriter:
    def __init__(self, dataframes: Union[pd.DataFrame, Dict[pd.DataFrame], List[pd.DataFrame]],
                 filename: PathLike, overwrite: bool = False, float_format: str = '%.6f',
                 sep: str = '\t', na_rep: str = '<NA>', force_loop: bool = False):
        self.overwrite = overwrite
        self.filename = filename
        self.dataframes = dataframes
        self.float_format = float_format
        self.sep = sep
        self.na_rep = na_rep
        self.force_loop = force_loop
        self.buffer = TextBuffer()
        self.write_star_file()

    @property
    def dataframes(self):
        """
        Ordered dictionary of pandas dataframes
        df.name defines the data block name
        """
        return self._dataframes

    @dataframes.setter
    def dataframes(self, dataframes: Union[pd.DataFrame, Dict[pd.DataFrame], List[pd.DataFrame]]):
        if isinstance(dataframes, pd.DataFrame):
            self._dataframes = self.coerce_dataframe(dataframes)
        elif isinstance(dataframes, dict):
            self._dataframes = self.coerce_dict(dataframes)
        elif isinstance(dataframes, list):
            self._dataframes = self.coerce_list(dataframes)
        else:
            raise ValueError(f'Expected a DataFrame, Dict or List object, got {type(dataframes)}')

    @staticmethod
    def coerce_dataframe(df: pd.DataFrame):
        name = getattr(df, 'name', '')
        if name != '':
            name = 0
        return {name: df}

    @staticmethod
    def coerce_dict(dfs: Dict[str, pd.DataFrame]):
        """
        This method ensures that dataframe names are updated based on dict keys
        """
        for key, df in dfs.items():
            df.name = str(key)
        return dfs

    def coerce_list(self, dfs: List[pd.DataFrame]):
        """
        This method coerces a list of DataFrames into a dict
        """
        return self.coerce_dict(OrderedDict([(idx, df) for idx, df in enumerate(dfs)]))

    @property
    def filename(self):
        return self._filename

    @filename.setter
    def filename(self, filename: Union[Path, str]):
        self._filename = Path(filename)
        if not self.file_writable:
            raise FileExistsError('to overwrite an existing file set overwrite=True')

    @property
    def file_exists(self):
        return self.filename.exists()

    @property
    def file_writable(self):
        if self.overwrite or (not self.file_exists):
            return True
        else:
            return False

    def write_package_info(self):
        date = datetime.now().strftime('%d/%m/%Y')
        time = datetime.now().strftime('%H:%M:%S')
        line = f'Created by the starfile Python package (version {__version__}) at {time} on' \
               f' {date}'
        self.buffer.add_comment(line)
        self.buffer.add_blank_lines(1)
        self.buffer.write_as_new_file_and_clear(self.filename)

    def write_star_file(self, filename: str = None):
        self.write_package_info()

        for _, df in self.dataframes.items():
            self.write_block(df)

        self.buffer.add_blank_line()
        self.buffer.append_to_file_and_clear(self.filename)

    def write_loopheader(self, df: pd.DataFrame):
        self.buffer.add_line('loop_')
        lines = [f'_{column_name} #{idx}' for idx, column_name in enumerate(df.columns, 1)]

        for line in lines:
            self.buffer.add_line(line)

        self.buffer.append_to_file_and_clear(self.filename)

    @staticmethod
    def get_block_name(df: pd.DataFrame):
        return 'data_' + getattr(df, 'name', '')

    def add_block_name_to_buffer(self, df: pd.DataFrame):
        self.buffer.add_line(self.get_block_name(df))
        self.buffer.add_blank_lines(1)
        self.buffer.append_to_file_and_clear(self.filename)

    def write_block(self, df: pd.DataFrame):
        self.add_block_name_to_buffer(df)

        if (df.shape[0] == 1) and not self.force_loop:
            self._write_simple_block(df)
        elif (df.shape[0] > 1) or self.force_loop:
            self._write_loop_block(df)
        self.buffer.add_blank_lines(2)
        self.buffer.append_to_file_and_clear(self.filename)

    def _write_simple_block(self, df: pd.DataFrame):
        lines = [f'_{column_name}\t\t\t{df[column_name].iloc[0]}'
                 for column_name in df.columns]
        for line in lines:
            self.buffer.add_line(line)
        self.buffer.append_to_file_and_clear(self.filename)

    def _write_loop_block(self, df: pd.DataFrame):
        self.write_loopheader(df)
        df.to_csv(self.filename, mode='a', sep=self.sep, header=False, index=False,
                  float_format=self.float_format, na_rep=self.na_rep)



from .constants import two_single_line_loop_blocks

import starfile
import pandas as pd


def test_round_trip_two_single_line_loop_blocks(tmp_path):
    # read
    expected = starfile.read(two_single_line_loop_blocks)

    # write
    output_file = tmp_path / 'two_single_line_loop_blocks.star'
    starfile.write(expected, output_file)

    # read
    star_after_round_trip = starfile.read(output_file)

    # assert
    for expected_df, actual_df in zip(expected.values(), star_after_round_trip.values()):
        pd.testing.assert_frame_equal(expected_df, actual_df)




from __future__ import annotations

from collections import OrderedDict
from io import StringIO

import pandas as pd
from pathlib import Path
from typing import TYPE_CHECKING, List, Union, Optional

from .utils import TextBuffer, TextCrawler

if TYPE_CHECKING:
    from os import PathLike


class StarParser:
    def __init__(self, filename: PathLike, read_n_blocks: Optional[int] = None):
        # set filename, with path checking
        self.filename = filename

        # initialise attributes for parsing
        self.text_buffer = TextBuffer()
        self.crawler = TextCrawler(self.filename)
        self.read_n_blocks = read_n_blocks
        self._dataframes = OrderedDict()
        self._current_dataframe_index = 0
        self._initialise_n_lines()

        # parse file
        self.parse_file()

    def parse_file(self):
        while self.crawler.current_line_number <= self.n_lines:
            if len(self.dataframes) == self.read_n_blocks:
                break

            elif self.crawler.current_line.startswith('data_'):
                self._parse_data_block()

            if not self.crawler.current_line.startswith('data_'):
                self.crawler.increment_line_number()

        self.dataframes_to_numeric()
        return

    def _parse_data_block(self):
        self.current_block_name = self._block_name_from_current_line()

        while self.crawler.current_line_number <= self.n_lines:
            self.crawler.increment_line_number()
            line = self.crawler.current_line

            if line.startswith('loop_'):
                self._parse_loop_block()
                return

            elif line.startswith('data_') or self.crawler.current_line_number == self.n_lines:
                self._parse_simple_block_from_buffer()
                return

            self.text_buffer.add_line(line)
        return

    def _parse_simple_block_from_buffer(self):
        data = self._clean_simple_block_in_buffer()

        df = self._cleaned_simple_block_to_dataframe(data)
        df.name = self._current_data_block_name
        self._add_dataframe(df)

        self.text_buffer.clear()

    def _parse_loop_block(self):
        self.crawler.increment_line_number()
        header = self._parse_loop_header()
        df = self._parse_loop_data()
        df.columns = header
        df.name = self._current_data_block_name
        self._add_dataframe(df)
        return

    @property
    def filename(self):
        return self._filename

    @filename.setter
    def filename(self, filename: Union[str, Path]):
        filename = Path(filename)
        if filename.exists():
            self._filename = filename
        else:
            raise FileNotFoundError

    @property
    def n_lines(self):
        return self._n_lines

    def _initialise_n_lines(self):
        self._n_lines = self.crawler.count_lines()

    @property
    def dataframes(self):
        return self._dataframes

    def _add_dataframe(self, df: pd.DataFrame):
        key = self._get_dataframe_key(df)
        self.dataframes[key] = df
        self._increment_dataframe_index()

    @property
    def current_block_name(self):
        return self._current_data_block_name

    @current_block_name.setter
    def current_block_name(self, name: str):
        self._current_data_block_name = name

    @property
    def current_dataframe_index(self):
        return self._current_dataframe_index

    def _increment_dataframe_index(self):
        self._current_dataframe_index += 1

    def _get_dataframe_key(self, df):
        name = df.name

        if name == '' or isinstance(name, int) or name in self.dataframes.keys():
            return self._current_dataframe_index
        else:
            return df.name

    def _clean_simple_block_in_buffer(self):
        clean_datablock = {}

        for line in self.text_buffer.buffer:
            if line == '' or line.startswith('#'):
                continue

            heading_name = self.heading_from_line(line)
            value = line.split()[1]
            clean_datablock[heading_name] = value

        return clean_datablock

    @staticmethod
    def _cleaned_simple_block_to_dataframe(data: dict):
        return pd.DataFrame(data, columns=data.keys(), index=[0])

    def _parse_loop_header(self) -> List[str]:
        self.text_buffer.clear()

        while self.crawler.current_line.startswith('_'):
            heading = self.heading_from_line(self.crawler.current_line)
            self.text_buffer.add_line(heading)
            self.crawler.increment_line_number()

        return self.text_buffer.buffer

    def _parse_loop_data(self) -> pd.DataFrame:
        self.text_buffer.clear()

        while self.crawler.current_line_number <= self.n_lines:
            current_line = self.crawler.current_line
            if current_line.startswith('data_'):
                break
            self.text_buffer.add_line(current_line)
            self.crawler.increment_line_number()

        df = pd.read_csv(StringIO(self.text_buffer.as_str()), delim_whitespace=True, header=None,
                         comment='#')
        return df

    def dataframes_to_numeric(self):
        """
        Converts strings in dataframes into numerical values where possible

        applying pd.to_numeric causes loss of 'name' attribute of DataFrame,
        need to extract name and reapply inline
        """
        for key, df in self.dataframes.items():
            name = getattr(df, 'name', None)
            self.dataframes[key] = df.apply(pd.to_numeric, errors='ignore')
            if name is not None:
                self.dataframes[key].name = name

    @staticmethod
    def _block_name_from_line(line: str):
        return line[5:]

    def _block_name_from_current_line(self):
        return self._block_name_from_line(self.crawler.current_line)

    @staticmethod
    def heading_from_line(line: str):
        return line.split()[0][1:]

    @property
    def first_dataframe(self):
        return self.dataframe_at_index(0)

    def dataframe_at_index(self, idx: int):
        return self.dataframes_as_list()[idx]

    def dataframes_as_list(self):
        return list(self.dataframes.values())







from os.path import join as join_path
from tempfile import TemporaryDirectory

import pandas as pd

from starfile.parser import StarParser
from starfile.writer import StarWriter

from .constants import loop_simple, postprocess, pipeline, rln31_style, optimiser_2d, optimiser_3d, sampling_2d, \
    sampling_3d, test_data_directory, test_df


def test_write_simple_block():
    s = StarParser(postprocess)
    output_file = test_data_directory / 'basic_block.star'
    StarWriter(s.dataframes, output_file, overwrite=True)
    assert output_file.exists()


def test_write_loop():
    s = StarParser(loop_simple)
    output_file = test_data_directory / 'loop_block.star'
    StarWriter(s.dataframes, output_file, overwrite=True)
    assert output_file.exists()


def test_write_multiblock():
    s = StarParser(postprocess)
    output_file = test_data_directory / 'multiblock.star'
    StarWriter(s.dataframes, output_file, overwrite=True)
    assert output_file.exists()


def test_from_single_dataframe():
    output_file = test_data_directory / 'from_df.star'

    StarWriter(test_df, output_file, overwrite=True)
    assert output_file.exists()

    s = StarParser(output_file)


def test_create_from_dataframes():
    dfs = [test_df, test_df]

    output_file = test_data_directory / 'from_list.star'
    StarWriter(dfs, output_file, overwrite=True)
    assert output_file.exists()

    s = StarParser(output_file)
    assert len(s.dataframes) == 2

def test_can_write_non_zero_indexed_one_row_dataframe():
    # see PR #13 - https://github.com/alisterburt/starfile/pull/13
    df = pd.DataFrame([[1,2,3]], columns=["A", "B", "C"])
    df.index += 1

    with TemporaryDirectory() as directory:
        filename = join_path(directory, "test.star")
        StarWriter(df, filename)
        with open(filename) as output_file:
            output = output_file.read()

    expected = (
        "_A\t\t\t1\n"
        "_B\t\t\t2\n"
        "_C\t\t\t3\n"
    )
    assert (expected in output)



try:
    from IPython.terminal.embed import InteractiveShellEmbed
    import click
    import matplotlib.pyplot as plt
    import seaborn as sns
except ImportError:
    deps = False
else:
    deps = True


if deps:
    from .functions import read, write

    @click.command()
    @click.argument('path', type=click.Path(exists=True, dir_okay=False, readable=True))
    @click.option('--read_n_blocks', type=int)
    @click.option('--always_dict', is_flag=True)
    def cli(path, read_n_blocks, always_dict):
        """
        Read a star file and open an ipython console to interactively inspect its contents
        """
        star = read(path, read_n_blocks, always_dict)

        banner = '''=== Starfile ===
    - access your data with `star`
    - write it out with `write(...)`
        '''
        sh = InteractiveShellEmbed(banner2=banner)
        sh.push('star')
        sh()
else:
    def cli():
        print('To use the command line utility, install with `pip install starfile[cli]`')



import time

import pandas as pd
import numpy as np
import pytest

from starfile.parser import StarParser
from .constants import loop_simple, postprocess, pipeline, rln31_style, optimiser_2d, optimiser_3d, \
    sampling_2d, \
    sampling_3d, single_line_middle_of_multiblock, single_line_end_of_multiblock, non_existant_file, \
    loop_simple_columns, two_single_line_loop_blocks, two_basic_blocks
from .utils import generate_large_star_file, remove_large_star_file, million_row_file


def test_instantiation():
    """
    Tests instantiation of the StarFile class
    """
    # instantiation with file which exists
    s = StarParser(loop_simple)

    # instantiation with non-existant file should fail
    assert non_existant_file.exists() is False
    with pytest.raises(FileNotFoundError):
        s = StarParser(non_existant_file)


def test_read_loop_block():
    """
    Check that loop block is parsed correctly, data has the correct shape
    """
    s = StarParser(loop_simple)

    # Check the output
    for idx, key in enumerate(s.dataframes.keys()):
        # Check that only one object is present
        assert idx < 1

        # get dataframe
        df = s.dataframes[key]
        assert isinstance(df, pd.DataFrame)

        # Check shape of dataframe
        assert df.shape == (16, 12)

        # check columns
        assert all(df.columns == loop_simple_columns)


def test_read_multiblock_file():
    """
    Check that multiblock STAR files such as postprocess RELION files
    parse properly
    """
    s = StarParser(postprocess)
    assert len(s.dataframes) == 3

    for key, df in s.dataframes.items():
        assert isinstance(df, pd.DataFrame)

    assert s.dataframes['general'].shape == (1, 6)
    assert all(['rlnFinalResolution', 'rlnBfactorUsedForSharpening', 'rlnUnfilteredMapHalf1',
                'rlnUnfilteredMapHalf2', 'rlnMaskName', 'rlnRandomiseFrom']
               == s.dataframes['general'].columns)
    assert s.dataframes['fsc'].shape == (49, 7)
    assert s.dataframes['guinier'].shape == (49, 3)


def test_read_pipeline():
    """
    Check that a pipeline.star file is parsed correctly
    """
    s = StarParser(pipeline)
    for key, df in s.dataframes.items():
        assert isinstance(df, pd.DataFrame)

    # Check that dataframes have the correct shapes
    assert s.dataframes['pipeline_general'].shape == (1, 1)
    assert s.dataframes['pipeline_processes'].shape == (31, 4)
    assert s.dataframes['pipeline_nodes'].shape == (74, 2)
    assert s.dataframes['pipeline_input_edges'].shape == (48, 2)
    assert s.dataframes['pipeline_output_edges'].shape == (72, 2)


def test_read_rln31():
    """
    Check that reading of RELION 3.1 style star files works properly
    """
    s = StarParser(rln31_style)

    for key, df in s.dataframes.items():
        assert isinstance(df, pd.DataFrame)

    assert isinstance(s.dataframes['block_1'], pd.DataFrame)
    assert isinstance(s.dataframes['block_2'], pd.DataFrame)
    assert isinstance(s.dataframes['block_3'], pd.DataFrame)


def test_read_n_blocks():
    """
    Check that passing read_n_blocks allows reading of only a specified
    number of data blocks from a star file
    """
    # test 1 block
    s = StarParser(postprocess, read_n_blocks=1)
    assert len(s.dataframes) == 1

    # test 2 blocks
    s = StarParser(postprocess, read_n_blocks=2)
    assert len(s.dataframes) == 2


def test_single_line_middle_of_multiblock():
    s = StarParser(single_line_middle_of_multiblock)
    assert len(s.dataframes) == 2


def test_single_line_end_of_multiblock():
    s = StarParser(single_line_end_of_multiblock)
    assert len(s.dataframes) == 2

    # iterate over dataframes, checking keys, names and shapes
    for idx, (key, df) in enumerate(s.dataframes.items()):
        assert df.name == 'block_3'
        if idx == 0:
            assert key == 'block_3'
            assert df.shape == (2, 5)
        if idx == 1:
            assert key == 1
            assert df.shape == (1, 5)


def test_read_optimiser_2d():
    s = StarParser(optimiser_2d)
    assert len(s.dataframes) == 1
    assert s.dataframes['optimiser_general'].shape == (1, 84)


def test_read_optimiser_3d():
    s = StarParser(optimiser_3d)
    assert len(s.dataframes) == 1
    assert s.dataframes['optimiser_general'].shape == (1, 84)


def test_read_sampling_2d():
    s = StarParser(sampling_2d)
    assert len(s.dataframes) == 1
    assert s.dataframes['sampling_general'].shape == (1, 12)


def test_read_sampling_3d():
    s = StarParser(sampling_3d)
    assert len(s.dataframes) == 2
    assert s.dataframes['sampling_general'].shape == (1, 15)
    assert s.dataframes['sampling_directions'].shape == (192, 2)


def test_df_as_list():
    s = StarParser(sampling_3d)
    assert isinstance(s.dataframes_as_list(), list)
    assert len(s.dataframes_as_list()) == 2


def test_first_dataframe():
    s = StarParser(sampling_3d)
    assert isinstance(s.first_dataframe, pd.DataFrame)


def test_parsing_speed():
    generate_large_star_file()
    start = time.time()
    s = StarParser(million_row_file)
    end = time.time()
    remove_large_star_file()

    # Check that execution takes less than a second
    assert end - start < 1


def test_two_single_line_loop_blocks():
    parser = StarParser(two_single_line_loop_blocks)
    assert len(parser.dataframes) == 2

    np.testing.assert_array_equal(
        parser.dataframes['block_0'].columns, [f'val{i}' for i in (1, 2, 3)]
    )
    assert parser.dataframes['block_0'].shape == (1, 3)

    np.testing.assert_array_equal(
        parser.dataframes['block_1'].columns, [f'col{i}' for i in (1, 2, 3)]
    )
    assert parser.dataframes['block_1'].shape == (1, 3)


def test_two_basic_blocks():
    parser = StarParser(two_basic_blocks)
    assert len(parser.dataframes) == 2
    for df in parser.dataframes.values():
        assert df.shape == (1, 3)
